@@ -8,6 +8,9 @@ class SQLGeneral(Exception):
     '''General error to prevent exposing SQL server specific error messages.'''
     pass
 
+class SQLInvalidInsertFormat(Exception):
+    '''Exception for value in an incorrect format such as a date string of mm/dd/yyyy.'''
+
 class SQLTableDoesNotExist(Exception):
     '''Exception for SQL table that does not exist.'''
     pass
@@ -28,21 +31,17 @@ class SQLInsufficientColumnSize(Exception):
     '''Exception for insufficient column size to insert a numeric or string.'''
     pass
 
-class SQLUndefinedColumn(Exception):
-    '''Exception for undefined SQL column.'''
-    pass
-
 class SQLInvalidLengthObjectName(Exception):
     '''Exception for an SQL object name that is too long.'''
     pass
 
+class SQLInvalidDataType(Exception):
+    '''Exception for invalid SQL data type.'''
+
 class SQLObjectAdjustment(Warning):
     '''Warning for creating of modifying an SQL object'''
     pass
 
 class DataframeUndefinedColumn(Exception):
     '''Exception for undefined dataframe column.'''
-    pass
-
-class DataframeUndefinedBestType(Warning):
-    '''Warning for undefined best data type of dataframe column given an SQL data type.'''
\ No newline at end of file
+    pass
\ No newline at end of file
